/* dfmax.c
* Semi-Exhaustive Greedy Independent Set
* from Matula and Johri
*
* written October 22, 1983 by DSJ
* modified February 20, 1987 for checkpointing
* modified August 1988 to find maximum independent sets
* modified April 1993 for new data structures, bounding
* modified September 1993 for dimacs .b input
* modified December 1993 to modify output to use as machine
*    benchmark for DIMACS Challenge Volume by Michael Trick
*
*  Modified timing routines to work in windows environment
*     on 6/21/2002 by Chirag Patel cmp196@psu.edu | chirag71@hotmail.com
*  Modified program to read ascii instance file instead of binary so that program work in Windows
*     on 8/12/2002 by Chirag Patel cmp196@psu.edu | chirag71@hotmail.com
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h> 

//added for win32
#include <windows.h>
#include <time.h>
#include <iostream.h>

#define INT 32		/* computer word size */
#define CHARBITS 8
#define edge(x,y) (bitmap[y/CHARBITS][x] & (1<<(y%CHARBITS)))

#define NMAX 2400		/* maximum number of vertices handles */
#define MAX_NR_VERTICES		2400	/* = NMAX */
#define MAX_NR_VERTICESdiv8	300	/* = NMAX/8 */
#define BOOL	char
#define MAX_PREAMBLE 10000
#define REORDER


 /* graph input parameters */

int Nr_vert, Nr_edges;
BOOL Bitmap[MAX_NR_VERTICES][MAX_NR_VERTICESdiv8];
static char Preamble[MAX_PREAMBLE];
char masks[ 8 ] = { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };




unsigned char bitmap[NMAX/CHARBITS+1][NMAX+1];
int N;			/* number of vertices in graph */
int mattype;
double dens, maxdist, dens1, dens2;
int chrom_num;		/* for fixed-chrom-num graphs */
double x_coord[NMAX+1], y_coord[NMAX+1];
int width;

int setlim;	/* size at which exhaustive search begins */

int vertex[NMAX];	/* array in which vertices reside and are moved */
int degree[NMAX];	/* vertex degrees with resp. to uncolored vertices */

int set[NMAX];		/* non-zero entries are members of current set */

int bestset[NMAX];	/* non-zero entries constitute currently best ind. set */
int bestsize;		/* size of current best set */
double bignum;		/* 2**31 */

void check();		/* checks to see that chosen color class is ind. set */
void readgraph();	/* reads input graph in bitmapform */


unsigned int startKernelTime;
unsigned int startUserTime;
time_t startRealTime;

unsigned int endKernelTime;
unsigned int endUserTime;
time_t endRealTime;

FILE *fopen();
char inputname[40];
int i;


int maxind(register int top, register int goal, int *array, register int depth)
//register int top, goal,depth;
//int *array;
{
	int newarray[NMAX];
	int i,w,z;
	//int u,v;
	int best, restbest, newgoal;
	//	unsigned *bitloc;
	int *pnew, *pold;
	int canthrow;

	if (top <= 1) {
		if (top == 0) depth--;
		if (depth > bestsize) {
			bestsize = depth;
			if (top == 1) set[bestsize] = array[top];
			for (i=1;i<=bestsize;i++) bestset[i] = set[i];
			check();
/*			printf("Size = %2d found\n",bestsize);*/
		}
		return(top);
	}
	best = 1;
	newgoal = goal-1;
	if (newgoal <= 1) newgoal = 1;
	for (i = top; i >= goal; i--) {
		pnew = newarray;
		w = array[i];
		set[depth] = w;
		canthrow = i - goal;
		pold = array+1;
		while (pold<array+i) {
			z = *pold++;
			if (edge(z,w)) {
				 *++pnew = z;
			} else {
			    if (canthrow == 0) goto breakout;
			    canthrow--;
			}
		}
		restbest = maxind(pnew-newarray,newgoal,newarray,depth+1);
		if (restbest >= newgoal) {
			best = newgoal = restbest+1;
			goal = best+1;
		}
		if (top == N) {
		
/*			printf("N = %3d best = %2d %8.2f (user)  %8.2f (sys)  %8.2f (real)\n",i,best, utime, stime, rtime);*/
		}
    breakout:;
	}
	return(best);
}

BOOL get_edge(int i, int j )
//int i,j;
{
	int byte, bit;
	char mask;
	int k;

	if (i<j) {
		k = i;
		i = j;
		j = k;
	}

	bit  = 7-(j & 0x00000007);
	byte = j >> 3;

	mask = masks[bit];
	return( (Bitmap[i][byte] & mask)==mask );
}

int get_params()
                      /* getting Nr_vert and Nr_edge from the preamble string,
			containing Dimacs format "p ??? num num" */
{
	char c, *tmp;
	char * pp = Preamble;
	int stop = 0;
	tmp = (char *)calloc(100, sizeof(char));

	Nr_vert = Nr_edges = 0;

	while (!stop && (c = *pp++) != '\0'){
		switch (c)
		  {
			case 'c':
			  while ((c = *pp++) != '\n' && c != '\0');
			  break;

			case 'p':
			  sscanf(pp, "%s %d %d\n", tmp, &Nr_vert, &Nr_edges);
			  stop = 1;
			  break;

			default:
			  break;
		  }
	}

	free(tmp);

	if (Nr_vert == 0 || Nr_edges == 0)
	  return 0;  /* error */
	else
	  return 1;

}

//
// **************** added 8/4/02****************************
//

void set_edge( register int i, register int j, BOOL x )
{
	register int byte, bit, mask;
	
	bit  = 7 - (j & 0x00000007);
	byte = j >> 3;
	
	mask = masks[bit];
	if ( x == 1 ) Bitmap[i][byte] |= mask;
	else  Bitmap[i][byte] &= ~mask;
}


void read_graph_DIMACS_ascii( FILE *fp)
{

	int c, oc;
	char * pp = Preamble;
	int i,j;
	

	for(oc = '\0' ;(c = fgetc(fp)) != EOF && (oc != '\n' || c != 'e')
		; oc = *pp++ = c);
 
	ungetc(c, fp); 
	*pp = '\0';
	get_params();

	while ((c = fgetc(fp)) != EOF){
		switch (c)
		  {
			case 'e':
			  if (!fscanf(fp, "%d %d", &i, &j))
				{ printf("ERROR: corrupted inputfile\n"); exit(10); }
			  
			  if (i > j) 
				set_edge(i, j, 1);
			  else
				set_edge(j, i, 1);
			  break;
			  
			case '\n':
			  
			default:
			  break;
		  }
	}
	
	fclose(fp);

	//added 8/04/02
	N = Nr_vert;

   for (i = 0; i < N; i++){
       for (j = 0; j < N; j++){
		   if (get_edge(i, j)){
			bitmap[(j+1)/CHARBITS][i+1] |= (1 << ((j+1) % CHARBITS));
		   
		   }
	   }
	}
   // added above
}

// *************************************************

void readgraph(FILE *fp)
//FILE *fp;
{
   int i,j;
   int length = 0;
	//unsigned temp;

	if (!fscanf(fp, "%d\n", &length))
	  { printf("ERROR: Corrupted preamble.\n"); exit(10); }

	if(length >= MAX_PREAMBLE)
	  { printf("ERROR: Too long preamble.\n"); exit(10); }

	fread(Preamble, 1, length, fp);
	Preamble[length] = '\0';

	if (!get_params())
		  { printf("ERROR: Corrupted preamble.\n"); exit(10); }

	if (Nr_vert >NMAX) {
		printf("Too many vertices! Recompile with NMAX > %d\n",
			Nr_vert);
		exit(0);
	}

	for ( i = 0
		 ; i < Nr_vert && fread(Bitmap[i], 1, (int)((i + 8)/8), fp)
		 ; i++ );

	fclose(fp);

   N = Nr_vert;

   for (i = 0; i < N; i++){
       for (j = 0; j < N; j++){
		   if (get_edge(i, j)){
			bitmap[(j+1)/CHARBITS][i+1] |= (1 << ((j+1) % CHARBITS));
		  
		   }
	   }
	}
}



void check ()
{
	/* checks that the vertices with color K are an independent set */
	register int i,j;

	for (i=1;i<=bestsize;i++)
		for (j=i+1;j<=bestsize;j++)
			if (!edge(bestset[i],bestset[j]))
				printf ("Edge in set\n");
}

///////////////////////////////////////////////////////
//
// get kernel, user and real time elapsed for NT/2000
// OR just cpu time and real time elapsed for W9x
//
// getStartTimes function is provided by Chirag Patel
// email:(cmp196@psu.edu | chirag71@hotmail.com)
//
///////////////////////////////////////////////////////
void getStartTimes()

{ //kernel, user and real time

	//if windows NT or higher
	if(GetVersion() < 0x80000000){
	FILETIME ftCreation;
	FILETIME ftExit;
	FILETIME ftKernel;
	FILETIME ftUser;
	ULARGE_INTEGER liKernel;
	ULARGE_INTEGER liUser;

	GetProcessTimes(GetCurrentProcess(), &ftCreation, &ftExit, &ftKernel, &ftUser);
	liKernel.LowPart = ftKernel.dwLowDateTime;
	liKernel.HighPart = ftKernel.dwHighDateTime;
	liUser.LowPart = ftUser.dwLowDateTime;
	liUser.HighPart = ftUser.dwHighDateTime;

	liKernel.QuadPart /= 10000000;
	liUser.QuadPart /= 10000000;

	startKernelTime = liKernel.LowPart;
	startUserTime = liUser.LowPart;
	}
	else{//if it is windows 95/98 or Win32s (Windows 3.x)

		startKernelTime = 0;
		startUserTime = clock()/CLOCKS_PER_SEC; //USER time is total CPU time for Windows 9x;

	}

	time(&startRealTime);

}

///////////////////////////////////////////////////////
//
// get kernel, user and real time elapsed for NT/2000
// OR just cpu time and real time elapsed for W9x
//
// getEndTimes function is provided by Chirag Patel
// email:(cmp196@psu.edu | chirag71@hotmail.com)
//
///////////////////////////////////////////////////////
void getEndTimes()

{ //kernel, user and real time
	//part of the  code for this routine came from Luc Kumps at kumps@attglobal.net
	//code was posted on web at http://members.tripod.com/~POTM/REMNANTS/timecode.txt

	// OR visit http://members.tripod.com/~POTM/

	//Windows version detection was added by Chirag Patel, cmp196@psu.edu | chirag71@hotmail.com

	//if windows NT or higher
	if(GetVersion() < 0x80000000){
	FILETIME ftCreation;
	FILETIME ftExit;
	FILETIME ftKernel;
	FILETIME ftUser;
	ULARGE_INTEGER liKernel;
	ULARGE_INTEGER liUser;

	GetProcessTimes(GetCurrentProcess(), &ftCreation, &ftExit, &ftKernel, &ftUser);
	liKernel.LowPart = ftKernel.dwLowDateTime;
	liKernel.HighPart = ftKernel.dwHighDateTime;
	liUser.LowPart = ftUser.dwLowDateTime;
	liUser.HighPart = ftUser.dwHighDateTime;

	liKernel.QuadPart /= 10000000;
	liUser.QuadPart /= 10000000;

	endKernelTime = liKernel.LowPart;
	endUserTime = liUser.LowPart;
	}
	else{//if it is windows 95/98 or Win32s (Windows 3.x)

		endKernelTime = 0;
		endUserTime = clock()/CLOCKS_PER_SEC; //USER time is total CPU time for Windows 9x;

	}

	time(&endRealTime);

}

main (int argc,char *argv[])
//int argc;
//char *argv[];
{
	int i,j,cand,newcand,dmax;
	FILE *inputfile;

	setbuf(stdout,NULL);

	

	/* Calculate bignum */
	bignum = (float)1024*1024;
	bignum *= 2048;

	/* read input */
	if (argc < 2) {
		printf("Usage: dfmax <filename> [setlim]\n");
		exit(1);
	}
	strcpy(inputname,argv[1]);
	if ((inputfile = fopen(argv[1], "r")) == NULL) {
		printf("Input graph does not exist");
		exit(1);
	}
	//readgraph(inputfile);
	read_graph_DIMACS_ascii(inputfile); //added 8/4/02




	getStartTimes();


	/* initialize algorithmic parameters */
	setlim = 1;
	if (argc > 2) setlim = atoi(argv[2]);
	printf("DFMAX(%s)\n",argv[1]);



#ifdef REORDER

	for (i=1;i<=N;i++) {
		degree[i] = 0;
		for (j=1;j<=N;j++)
			if (!edge(i,j)) degree[i]++;
	}

	dmax = -1;
	for (i=1;i<=N;i++)
		if (degree[i] > dmax) {
			dmax = degree[i];
			cand = i;
		}
	vertex[N] = cand;

	for (j=N-1;j>=1;j--) {
		degree[cand] = -9;
		dmax = -1;
		for (i=1;i<=N;i++) {
			if (!edge(cand,i)) degree[i]--;
			if (degree[i] > dmax) {
				dmax = degree[i];
				newcand = i;
			}
		}
		vertex[j] = cand = newcand;
	}
#else
	for (j=1;j<=N;j++) vertex[j] = j;
#endif

	bestsize = 0;
	bestsize = maxind(N,setlim,vertex,1);


	

	getEndTimes();

	cout<<"(sys) "<<endKernelTime-startKernelTime<< " (user) "<< endUserTime-startUserTime<<" (Real) "<< endRealTime-startRealTime<<endl;
	printf ("Best:");
	for (i=1; i<=bestsize; i++) {
	    printf (" %d",bestset[i]);
	}
	printf ("\n\n");

	return(0);

}









